#include <jacket.h>

#define BLOCK_X  16
#define BLOCK_Y  16

#define SMEM(x,y)  smem[(x)+1][(y)+1]
#define IN(x,y)    d_in[(y)*nx + (x)]

static __global__ void kernel(int nx, int ny, float *d_out, float *d_in)
{
    int tx = threadIdx.x, ty = threadIdx.y;

    // guards: is at boundary?
    bool is_x_top = (tx == 0), is_x_bot = (tx == BLOCK_X-1);
    bool is_y_top = (ty == 0), is_y_bot = (ty == BLOCK_Y-1);

    __shared__ float smem[BLOCK_X+2][BLOCK_Y+2];
    // clear out shared memory (zero padding)
    if (is_x_top)           SMEM(tx-1, ty  ) = 0;
    else if (is_x_bot)      SMEM(tx+1, ty  ) = 0;
    if (is_y_top) {         SMEM(tx  , ty-1) = 0;
        if (is_x_top)       SMEM(tx-1, ty-1) = 0;
        else if (is_x_bot)  SMEM(tx+1, ty-1) = 0;
    } else if (is_y_bot) {  SMEM(tx  , ty+1) = 0;
        if (is_x_top)       SMEM(tx-1, ty+1) = 0;
        else if (is_x_bot)  SMEM(tx+1, ty+1) = 0;
    }

    // guards: is at boundary and still more image?
    int x = blockIdx.x * blockDim.x + tx;
    int y = blockIdx.y * blockDim.y + ty;
    is_x_top &= (x > 0); is_x_bot &= (x < nx - 1);
    is_y_top &= (y > 0); is_y_bot &= (y < ny - 1);

    // each thread pulls from image
                            SMEM(tx  , ty  ) = IN(x  , y  ); // self
    if (is_x_top)           SMEM(tx-1, ty  ) = IN(x-1, y  );
    else if (is_x_bot)      SMEM(tx+1, ty  ) = IN(x+1, y  );
    if (is_y_top) {         SMEM(tx  , ty-1) = IN(x  , y-1);
        if (is_x_top)       SMEM(tx-1, ty-1) = IN(x-1, y-1);
        else if (is_x_bot)  SMEM(tx+1, ty-1) = IN(x+1, y-1);
    } else if (is_y_bot) {  SMEM(tx  , ty+1) = IN(x  , y+1);
        if (is_x_top)       SMEM(tx-1, ty+1) = IN(x-1, y+1);
        else if (is_x_bot)  SMEM(tx+1, ty+1) = IN(x+1, y+1);
    }
    __syncthreads();

    // pull my window from shared memory
    float v[9] = { SMEM(tx-1, ty-1), SMEM(tx  , ty-1), SMEM(tx+1, ty-1),
                   SMEM(tx-1, ty  ), SMEM(tx  , ty  ), SMEM(tx+1, ty  ),
                   SMEM(tx-1, ty+1), SMEM(tx  , ty+1), SMEM(tx+1, ty+1) };    

    // bubble-sort
    for (int i = 0; i < 9; i++) {
        for (int j = i + 1; j < 9; j++) {
            if (v[i] > v[j]) { // swap?
                float tmp = v[i];
                v[i] = v[j];
                v[j] = tmp;
            }
        }
    }
    // pick the middle one
    d_out[y*nx + x] = v[4];
}


err_t jktFunction(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
    mxArray *m_in = prhs[0];
    const mwSize *dims;
    int ndims = jkt_dims(m_in, &dims);

    mxArray *m_out = plhs[0] = jkt_new_array(ndims, dims, mxSINGLE_CLASS, false);

    // get pointers to input and output
    float *d_in, *d_out;
    TRY(jkt_mem((void **)&d_in, m_in));
    TRY(jkt_mem((void **)&d_out, m_out));

    // launch CUDA kernel
    dim3 blocks(dims[0]/BLOCK_X, dims[1]/BLOCK_Y);
    dim3 threads(BLOCK_X,BLOCK_Y);
    kernel<<<blocks,threads>>>(dims[0], dims[1], d_out, d_in);

    return errNone;
}
